/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { Resource, ResourceUri, getUrl, hasLink } from '@ngxp/rest';
import { isEqual, isNil, isNull } from 'lodash-es';
import { BehaviorSubject, Observable, combineLatest, filter, first, map, startWith, tap } from 'rxjs';
import { isNotNull, isNotUndefined } from '../tech.util';
import { CreateResourceData, ListItemResource, ListResourceServiceConfig } from './resource.model';
import { ResourceRepository } from './resource.repository';
import { mapToFirst, mapToResource } from './resource.rxjs.operator';
import {
  ListResource,
  StateResource,
  createEmptyStateResource,
  createStateResource,
  doIfLoadingRequired,
  getEmbeddedResources,
  isInvalidResourceCombination,
  isLoadingRequired,
  isStateResoureStable,
} from './resource.util';

/**
 * B = Type of baseresource
 * T = Type of listresource
 * I = Type of items in listresource
 */
export class ResourceListService<B extends Resource, T extends ListResource, I extends ListItemResource> {
  readonly nextLink: string = 'next';
  readonly prevLink: string = 'prev';

  readonly listResource: BehaviorSubject<StateResource<T>> = new BehaviorSubject(createEmptyStateResource());

  readonly selectedResource: BehaviorSubject<StateResource<I>> = new BehaviorSubject(createEmptyStateResource());

  baseResource: B = null;
  linkedResources = {};

  constructor(
    private config: ListResourceServiceConfig<B>,
    private repository: ResourceRepository,
  ) {}

  public getList(): Observable<StateResource<T>> {
    return combineLatest([this.listResource.asObservable(), this.getConfigResource()]).pipe(
      tap(([stateResource, configResource]) => {
        this.handleChanges(stateResource, configResource);
      }),
      filter(([stateResource]) => !isInvalidResourceCombination(stateResource, this.baseResource)),
      mapToFirst<T, B>(),
      startWith(createEmptyStateResource<T>(true)),
    );
  }

  private getConfigResource(): Observable<B> {
    return this.config.baseResource.pipe(filter(isStateResoureStable<B>), mapToResource<B>());
  }

  handleChanges(stateResource: StateResource<T>, configResource: B): void {
    if (!isEqual(this.baseResource, configResource)) {
      this.handleConfigResourceChanges(configResource);
    } else if (this.shouldLoadResource(stateResource, configResource)) {
      this.loadListResource(configResource, this.config.listLinkRel);
    }
  }

  handleConfigResourceChanges(newConfigResource: B): void {
    this.baseResource = newConfigResource;
    if (this.hasListLinkRel() && isStateResoureStable(this.listResource.value)) {
      this.loadListResource(this.baseResource, this.config.listLinkRel);
    } else if (!this.hasListLinkRel() && isStateResoureStable(this.listResource.value)) {
      this.clearCurrentListResource();
    }
  }

  private hasListLinkRel(): boolean {
    return hasLink(this.baseResource, this.config.listLinkRel);
  }

  shouldLoadResource(stateResource: StateResource<T>, configResource: B): boolean {
    return isNotNull(configResource) && isLoadingRequired(stateResource) && this.hasListLinkRel();
  }

  private clearCurrentListResource(): void {
    this.listResource.next(createEmptyStateResource());
  }

  public create(toCreate: unknown): Observable<Resource> {
    this.verifyBeforeCreation();
    return this.repository.createResource(this.buildCreateResourceData(toCreate, this.config.createLinkRel));
  }

  private verifyBeforeCreation(): void {
    this.verifyValidListResource();
    this.throwErrorOn(!this.isCreateLinkPresent(), 'No creation link exists.');
  }

  private buildCreateResourceData(toCreate: any, linkRel: string): CreateResourceData<T> {
    return {
      resource: this.getListResource(),
      linkRel,
      toCreate,
    };
  }

  private isCreateLinkPresent(): boolean {
    return this.hasLinkRel(this.config.createLinkRel);
  }

  public select(uri: ResourceUri): void {
    this.setSelectedResourceLoading();
    this.repository
      .getResource(uri)
      .pipe(first())
      .subscribe((loadedResource) => {
        this.selectedResource.next(createStateResource(<I>loadedResource));
      });
  }

  verifyBeforeSelection(uri: ResourceUri): void {
    this.verifyValidListResource();
    this.throwErrorOn(!this.existsUriInList(uri), 'No entry match with given uri.');
  }

  private verifyValidListResource(): void {
    this.throwErrorOn(isNull(this.getListResource()), 'No list resource available.');
  }

  setSelectedResourceLoading(): void {
    this.selectedResource.next({ ...this.selectedResource.value, loading: true });
  }

  existsUriInList(uri: ResourceUri): boolean {
    const listResources: Resource[] = getEmbeddedResources(this.listResource.value, this.config.listLinkRel);

    return isNotUndefined(listResources.find((resource) => getUrl(resource) === uri));
  }

  public unselect(): void {
    this.selectedResource.next(createEmptyStateResource());
  }

  public getSelected(): Observable<StateResource<Resource>> {
    return this.selectedResource.asObservable();
  }

  public refresh(): void {
    this.listResource.next({ ...this.listResource.value, reload: true });
  }

  public prev(): void {
    this.throwErrorOn(!this.hasPrevLink(), 'There is no previous page.');
    this.loadListResource(this.getListResource(), this.prevLink);
  }

  public next(): void {
    this.throwErrorOn(!this.hasNextLink(), 'There is no next page.');
    this.loadListResource(this.getListResource(), this.nextLink);
  }

  loadListResource(resource: B | T, linkRel: string): void {
    this.setStateResourceLoading();
    this.repository
      .getListResource(resource, linkRel)
      .pipe(first())
      .subscribe((loadedListResource: T) => this.updateListResource(loadedListResource));
  }

  setStateResourceLoading(): void {
    this.listResource.next(createEmptyStateResource(true));
  }

  updateListResource(listResource: T): void {
    this.listResource.next(createStateResource(listResource));
  }

  public getLinkedResource<F extends Resource>(resourceUri: ResourceUri): Observable<StateResource<F>> {
    this.createIfNotExist(resourceUri);
    this.loadLinkedResourceIfRequired(resourceUri);
    return this.linkedResources[resourceUri].asObservable();
  }

  createIfNotExist(uri: ResourceUri): void {
    if (isNil(this.linkedResources[uri])) this.linkedResources[uri] = new BehaviorSubject(createEmptyStateResource());
  }

  loadLinkedResourceIfRequired(resourceUri: ResourceUri): void {
    doIfLoadingRequired(this.linkedResources[resourceUri].value, () => {
      this.setLinkedResourceLoading(resourceUri);
      this.loadLinkedResoure(resourceUri);
    });
  }

  loadLinkedResoure(resourceUri: ResourceUri): void {
    this.repository
      .getResource(resourceUri)
      .pipe(first())
      .subscribe((loadedResource: any) => this.linkedResources[resourceUri].next(createStateResource(loadedResource)));
  }

  setLinkedResourceLoading(resourceUri: ResourceUri): void {
    this.linkedResources[resourceUri].next(createEmptyStateResource(true));
  }

  private throwErrorOn(condition: boolean, errorMsg: string): void {
    if (condition) throw Error(errorMsg);
  }

  public hasMore(): boolean {
    return this.hasNextLink();
  }

  private hasNextLink(): boolean {
    return this.hasLinkRel(this.nextLink);
  }

  public isFirst(): boolean {
    return !this.hasPrevLink();
  }

  private hasPrevLink(): boolean {
    return this.hasLinkRel(this.prevLink);
  }

  private hasLinkRel(linkRel: string): boolean {
    return hasLink(this.getListResource(), linkRel);
  }

  private getListResource(): T {
    return this.listResource.value.resource;
  }

  public getItems(): Observable<ListItemResource[]> {
    return this.getList().pipe(
      filter((listStateResource: StateResource<T>) => !listStateResource.loading),
      map((listStateResource: StateResource<T>) =>
        getEmbeddedResources<ListItemResource>(listStateResource, this.config.listResourceListLinkRel),
      ),
    );
  }
}
