/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.function.Supplier;
import java.util.stream.Stream;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.hateoas.Link;

import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.wiedervorlage.Wiedervorlage;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageTestFactory;

class CollectionModelBuilderTest {

	private final String HREF = "http://test";
	private final String REL = "rel";

	@Nested
	class TestBuildModel {

		@Test
		void shouldBuildModel() {
			var vorgang = VorgangHeaderTestFactory.create();

			var model = CollectionModelBuilder.fromEntities(List.of(vorgang)).buildModel();

			assertThat(model.getContent()).hasSize(1).first().usingRecursiveComparison().isEqualTo(vorgang);
		}
	}

	@Nested
	class TestAddLink {

		@Test
		void shouldAddLink() {
			var model = CollectionModelBuilder.fromEntities(List.of()).addLink(Link.of(HREF, REL)).buildModel();

			assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
		}
	}

	@Nested
	class TestIfMatch {

		@Nested
		class TestWithBooleanSupplier {

			@Test
			void shouldAddLink() {
				var model = CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> true).addLink(Link.of(HREF, REL)).buildModel();

				assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
			}

			@Test
			void shouldNotAddLink() {
				var model = CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> false).addLink(Link.of(HREF, REL)).buildModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}

		@Nested
		class TestWithPredicate {

			private final Stream<Wiedervorlage> wiedervorlageStream = Stream.of(WiedervorlageTestFactory.create());

			@Test
			void shouldAddLink() {
				var model = CollectionModelBuilder.fromEntities(wiedervorlageStream)
						.ifMatch(wiedervorlagen -> true)
						.addLink(Link.of(HREF, REL))
						.buildModel();

				assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
			}

			@Test
			void shouldNotAddLink() {
				var model = CollectionModelBuilder.fromEntities(wiedervorlageStream)
						.ifMatch(wiedervorlagen -> false)
						.addLink(Link.of(HREF, REL))
						.buildModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}

		@Nested
		class TestWithLinkSupplier {

			@Mock
			private Supplier<Link> linkSupplier;

			@Test
			void shouldAddLink() {
				when(linkSupplier.get()).thenReturn(Link.of(HREF, REL));

				var model = CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> true).addLink(linkSupplier).buildModel();

				assertThat(model.getLinks()).hasSize(1).first().extracting(Link::getHref, link -> link.getRel().value()).containsExactly(HREF, REL);
			}

			@Test
			void shouldNotCallLinkSupplier() {
				CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> false).addLink(linkSupplier).buildModel();

				verify(linkSupplier, never()).get();
			}

			@Test
			void shouldNotAddLink() {
				var model = CollectionModelBuilder.fromEntities(List.of()).ifMatch(() -> false).addLink(linkSupplier).buildModel();

				assertThat(model.getLinks()).isEmpty();
			}
		}
	}

}